# inspect_client.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2014 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import copy
import logging
import random

from builtins import object, range
from collections import defaultdict, namedtuple

import tornado

from tornado.gen import Return, maybe_future
from future.utils import native_str_to_bytes

import katcp.core

from katcp.core import (AttrDict, ensure_native_str, future_timeout_manager,
                        steal_docstring_from, until_any)

ic_logger = logging.getLogger("katcp.inspect_client")
RequestType = namedtuple('Request', 'name description timeout_hint')


class ExponentialRandomBackoff(object):

    def __init__(
            self, delay_initial=1., delay_max=90., exp_fac=3., randomicity=0.95):
        """Calculate random retry timeouts that increase exponentially

        Input Parameters
        ----------------

        delay_initial : float
            Initial base delay in seconds
        delay_max : float
            Maximum delay in seconds
        exp_fac : float
            Increase the base timeout by this factor for each failure
        randomicity : float
            Fraction of timeout that should be randomly calculated. If
            randomicity is 0, the base delay will always be used, if it is 1, a
            random value between 0 and the base delay is calculated, or the
            weighted average of the two for intermediate values.

        """
        self.delay_initial = delay_initial
        self.delay_max = delay_max
        self.exp_fac = exp_fac
        """Increase timeout by this factor for each consecutive failure"""
        self._base_delay = self.delay_initial
        self.randomicity = randomicity
        self._update_delay()

    def _validate_parameters(self):
        # Since a user could potentially change these parameters at any time and
        # since it is nice to give feedback at __init__ time as well, have a
        # single validation routine that can be called any place it matters
        assert 0 <= self.delay_initial <= self.delay_max
        assert 0 <= self.randomicity <= 1
        assert self.exp_fac > 1

    def _update_delay(self):
        self._validate_parameters()
        r = self.randomicity
        d = ((1 - r) + r*random.random())*self._base_delay
        self._delay = min(d, self.delay_max)

    def failed(self):
        """Call whenever an action has failed, grows delay exponentially

        After calling failed(), the `delay` property contains the next delay
        """
        try:
            self._validate_parameters()
            self._base_delay = min(
                self._base_delay * self.exp_fac,
                self.delay_max)
            self._update_delay()
        except Exception:
            ic_logger.exception(
                'Unhandled exception trying to calculate a retry timeout')

    def success(self):
        """Call whenever an action has succeeded, resets delay to minimum"""
        self._base_delay = self.delay_initial
        self._update_delay()

    @property
    def delay(self):
        return self._delay



class InspectingClientStateType(namedtuple(
        'InspectingClientStateType', 'connected synced model_changed data_synced')):
    """
    States tuple for the inspecting client. Fields, all bool:

    connected: bool
        TCP connection has been established with the server.
    synced: bool
        The inspecting client and the user that interfaces through the state change
        callback are all synchronised with the current device state. Also implies
        `connected = True` and `data_synced = True`.
    model_changed: bool
        The device has changed in some way, resulting in the device model being out of
        date.
    data_synced: bool
        The inspecting client's internal representation of the device is up to date,
        although state change user is not yet up to date.

    """
    __slots__ = []

    def __str__(self):
        def _boolstr(name):
            val = bool(getattr(self, name))
            return name if val else "!"+name
        return "[{} {} {} {}]".format(
            _boolstr('connected'),
            _boolstr('synced'),
            _boolstr('model_changed'),
            _boolstr('data_synced'))


class SyncError(Exception):
    """Raised if an error occurs during syncing with a device"""


class _InformHookDeviceClient(katcp.AsyncClient):
    """DeviceClient that adds inform hooks."""

    def __init__(self, *args, **kwargs):
        super(_InformHookDeviceClient, self).__init__(*args, **kwargs)
        self._inform_hooks = defaultdict(list)

    def hook_inform(self, inform_name, callback):
        """Hookup a function to be called when an inform is received.

        Useful for interface-changed and sensor-status informs.

        Parameters
        ----------
        inform_name : str
            The name of the inform.
        callback : function
            The function to be called.

        """
        # Do not hook the same callback multiple times
        if callback not in self._inform_hooks[inform_name]:
            self._inform_hooks[inform_name].append(callback)

    def handle_inform(self, msg):
        """Call callbacks on hooked informs followed by normal processing"""
        try:
            for func in self._inform_hooks.get(msg.name, []):
                func(msg)
        except Exception:
            self._logger.warning('Call to function "{0}" with message "{1}".'
                                 .format(func, msg), exc_info=True)
        super(_InformHookDeviceClient, self).handle_inform(msg)


class InspectingClientAsync(object):
    """Higher-level client that inspects a KATCP interface.

    Note: This class is not thread-safe at present, it should only be called
    from the ioloop.

    Note: always call stop() after start() and you are done with the container
    to make sure the container cleans up correctly.

    """
    sensor_factory = katcp.Sensor
    """Factory that produces a KATCP Sensor compatible instance.

    signature: sensor_factory(sensor_type,
                              name,
                              description,
                              units,
                              params)

    Should be set before calling connect()/start().

    """
    request_factory = RequestType
    """Factory that produces KATCP Request objects.

    signature: request_factory(name, description, timeout_hint), all parameters
    passed as kwargs

    Should be set before calling connect()/start().

    """

    # TODO (NM 2016-10-21) sync_timeout should be 5 seconds, but until we deal
    # with the thundering herd at startup when connecting to a large number of
    # clients concurrently in a single process, see Jira CB-1609
    sync_timeout = 20
    initial_resync_timeout = 5
    max_resync_timeout = 90

    def __init__(self, host, port, ioloop=None, initial_inspection=None,
                 auto_reconnect=True, logger=ic_logger):
        # TODO Consider optional 'name' parameter just to make logging clearer
        self._logger = logger
        self.resync_delay = None
        """Set to an ExponentialRandomBackoff instance in _state_loop"""
        if initial_inspection is None:
            initial_inspection = True
        self.initial_inspection = bool(initial_inspection)
        self._requests_index = {}
        self._sensors_index = {}
        self._sensor_object_cache = {}
        self._connected = katcp.core.AsyncEvent()
        self._disconnected = katcp.core.AsyncEvent()
        self._interface_changed = katcp.core.AsyncEvent()
        # Set the default behaviour for update.
        self._update_on_lookup = True
        self._cb_register = {}  # Register to hold the possible callbacks.
        self._running = False

        # Setup KATCP device.
        self.katcp_client = self.inform_hook_client_factory(
            host, port, auto_reconnect=auto_reconnect, logger=logger)
        self.ioloop = ioloop or tornado.ioloop.IOLoop.current()
        self.katcp_client.set_ioloop(ioloop)

        self.katcp_client.hook_inform('sensor-status',
                                      self._cb_inform_sensor_status)
        self.katcp_client.hook_inform('interface-changed',
                                      self._cb_inform_interface_change)
        self.katcp_client.hook_inform('device-changed',
                                      self._cb_inform_interface_change)
        # Hook a callback for/to deprecated informs.
        # _cb_inform_deprecated will log a message when one of these informs
        # are received.
        self.katcp_client.hook_inform('device-changed',
                                      self._cb_inform_deprecated)
        self.katcp_client.notify_connected = self._cb_connection_state

        # User callback function to be called on state changes
        self._state_cb = None
        self.valid_states = frozenset((
            InspectingClientStateType(
                connected=False, synced=False, model_changed=False, data_synced=False),
            InspectingClientStateType(
                connected=True, synced=False, model_changed=False, data_synced=False),
            InspectingClientStateType(
                connected=True, synced=False, model_changed=True, data_synced=True),
            InspectingClientStateType(
                connected=True, synced=False, model_changed=False, data_synced=True),
            InspectingClientStateType(
                connected=True, synced=True, model_changed=False, data_synced=True)))
        self._state = katcp.core.AsyncState(
            self.valid_states,
            InspectingClientStateType(
                connected=False, synced=False, model_changed=False, data_synced=False)
        )

    def inform_hook_client_factory(self, host, port, *args, **kwargs):
        """Return an instance of :class:`_InformHookDeviceClient` or similar

        Provided to ease testing. Dynamically overriding this method after instantiation
        but before start() is called allows for deep brain surgery. See
        :class:`katcp.fake_clients.TBD`

        """
        return _InformHookDeviceClient(host, port, *args, **kwargs)

    @property
    def state(self):
        """Current client state."""
        return self._state.state

    @property
    def sensors(self):
        """A list of known sensors."""
        return list(self._sensors_index.keys())

    @property
    def requests(self):
        """A list of possible requests."""
        return list(self._requests_index.keys())

    @property
    def connected(self):
        """Connection status."""
        return self.katcp_client.is_connected()

    @property
    def synced(self):
        """Boolean indicating if the device has been synchronised."""
        return self._state.state.synced

    @property
    def bind_address(self):
        return self.katcp_client.bind_address

    @property
    def bind_address_string(self):
        return self.katcp_client.bind_address_string

    @steal_docstring_from(_InformHookDeviceClient.preset_protocol_flags)
    def preset_protocol_flags(self, protocol_flags):
        self.katcp_client.preset_protocol_flags(protocol_flags)

    def set_ioloop(self, ioloop):
        self.katcp_client.set_ioloop(ioloop)

    def is_connected(self):
        """Connection status."""
        return self.katcp_client.is_connected()

    def until_connected(self, timeout=None):
        # TODO (NM) Perhaps misleading to say until_protocol here? For debugging it is
        # useful to know if the TCP connection is established even if nothing else has
        # happened yet. Also, won't match is_connected()
        return self.katcp_client.until_protocol(timeout=timeout)

    def until_synced(self, timeout=None):
        return self._state.until_state(
            InspectingClientStateType(connected=True, synced=True,
                                      model_changed=False, data_synced=True),
            timeout=timeout)

    def until_not_synced(self, timeout=None):
        unsynced_states = tuple(state for state in self.valid_states
                                if not state.synced)
        return self._state.until_state_in(*unsynced_states, timeout=timeout)

    def until_data_synced(self, timeout=None):
        data_synced_states = tuple(state for state in self.valid_states
                                   if state.data_synced)
        return self._state.until_state_in(*data_synced_states, timeout=timeout)

    def until_state(self, desired_state, timeout=None):
        """
        Wait until state is desired_state, InspectingClientStateType instance

        Returns a future

        """
        return self._state.until_state(desired_state, timeout=timeout)

    @tornado.gen.coroutine
    def connect(self, timeout=None):
        """Connect to KATCP interface, starting what is needed

        Parameters
        ----------
        timeout : float, None
            Time to wait until connected. No waiting if None.

        Raises
        ------

        :class:`tornado.gen.TimeoutError` if the connect timeout expires

        """
        # Start KATCP device client.
        assert not self._running
        maybe_timeout = future_timeout_manager(timeout)

        self._logger.debug('Starting katcp client')
        self.katcp_client.start()
        try:
            yield maybe_timeout(self.katcp_client.until_running())
            self._logger.debug('Katcp client running')
        except Exception as err:
            self._logger.error("Failed to start katcp_client: %s", err)
            self.katcp_client.stop()
            raise

        if timeout:
            yield maybe_timeout(self.katcp_client.until_connected())
        self._logger.debug('Katcp client connected')

        self._running = True
        self._state_loop()

    @katcp.core.log_coroutine_exceptions
    @tornado.gen.coroutine
    def _state_loop(self):
        # TODO (NM) Arrange for _running to be set to false and stopping the katcp client
        # if this loop exits
        self.resync_delay = ExponentialRandomBackoff(
            self.initial_resync_timeout, self.max_resync_timeout)

        is_connected = self.katcp_client.is_connected
        last_sync_failed = False
        while self._running:
            self._logger.debug('{}: Sending initial state'
                               .format(self.bind_address_string))
            yield self._send_state(connected=is_connected(), synced=False,
                                   model_changed=False, data_synced=False)
            try:
                yield self.katcp_client.until_connected()
                self._logger.debug('{}: Sending post-connected state'
                                   .format(self.bind_address_string))
                yield self._send_state(connected=is_connected(), synced=False,
                                       model_changed=False, data_synced=False)
                yield until_any(self.katcp_client.until_protocol(),
                                self._disconnected.until_set())
                # TODO NM 2016-10-21 The naming of this attribute is incredibly
                # confusing, we need to think about what "initial_inspection" is
                # and how it is supposed to be handled
                if self.initial_inspection:
                    if not is_connected():
                        continue
                    model_changes = yield self.inspect()
                    model_changed = bool(model_changes)
                    self._logger.debug('{}: Sending data-synced state'
                                       .format(self.bind_address_string))
                    yield self._send_state(
                        connected=True, synced=False,
                        model_changed=model_changed, data_synced=True,
                        model_changes=model_changes)
                else:
                    self.initial_inspection = True
                if not is_connected():
                    continue

                # TODO NM 2017-07-21 for CB-1466 (async dev handler) we need to
                # check if the device needs a re-sync after the state change
                # callback, since some proxy actions can result in the device
                # issuing an #interface-changed between _send_state() above and
                # here.

                # We waited for the previous _send_state call (and user callbacks) to
                # complete before we change the state to synced=True
                self._logger.debug('{}: Sending synced state'
                                   .format(self.bind_address_string))
                yield self._send_state(connected=True, synced=True,
                                       model_changed=False, data_synced=True)
                if last_sync_failed:
                    self._logger.warn(
                        'Succesfully resynced with {} after failure'
                        .format(self.bind_address_string))
                    last_sync_failed = False
                yield until_any(self._interface_changed.until_set(),
                                self._disconnected.until_set())
                self._logger.debug('in _state_loop: interface_changed=%s,'
                        ' is_connected=%s', self._interface_changed.is_set(),
                        self._disconnected.is_set())
                self._interface_changed.clear()
                continue
                # Next loop through should cause re-inspection and handle state updates
            except SyncError as e:
                last_sync_failed = True
                retry_wait_time = self.resync_delay.delay
                self.resync_delay.failed()
                self._logger.warn(
                    "Error syncing with device {}: {!s} 'Retrying in {}s."
                    .format(
                        self.bind_address_string,
                        e, retry_wait_time))
                yield katcp.core.until_later(retry_wait_time)
                # TODO (NM) Perhaps maintain count of unsuccessful attempts, and reconnect
                # if too many happen. Perhaps also integrate exponential-backoff stuff
                # here? Or outsource to a user-supplied class or callback?
                continue
            except Exception:
                last_sync_failed = True
                retry_wait_time = self.resync_delay.delay
                self.resync_delay.failed()
                self._logger.exception(
                    'Unhandled exception in client-sync loop. Triggering disconnect and '
                    'Retrying in {}s.'
                    .format(retry_wait_time))
                self.katcp_client.disconnect()
                yield katcp.core.until_later(retry_wait_time)
                continue
            else:
                self.resync_delay.success()
        try:
            self._logger.debug('{}: Sending final state - loop ended'
                               .format(self.bind_address_string))
            yield self._send_state(connected=False, synced=False,
                                   model_changed=False, data_synced=False)
        except Exception:
            self._logger.exception('Unhandled exception after client-sync loop ended.')

    @tornado.gen.coroutine
    def _send_state(self, connected, synced, model_changed, data_synced,
                    model_changes=None):
        # Should only be called from _state_loop()
        state = InspectingClientStateType(connected, synced, model_changed, data_synced)
        self._state.set_state(state)
        self._logger.debug('InspectingClient state changed to {0}'.format(state))

        if self._state_cb:
            yield maybe_future(self._state_cb(state, model_changes))
        # Make sure other callbacks in response to state change get to run before we
        # change state again
        yield tornado.gen.moment

    def set_state_callback(self, cb):
        """Set user callback for state changes

        Called as ``cb(state, model_changes)``

        where `state` is an :class:`InspectingClientStateType` instance, and
        `model_changes` is an :class:`~katcp.core.AttrDict`. The latter may
        contain keys ``requests`` and ``sensors`` to describe changes to
        requests or sensors respectively. These in turn have attributes
        ``added`` and ``removed`` which are sets of request/sensor names.
        Requests/sensors that have been modified will appear in both sets.

        .. warning::

            It is possible for `model_changes` to be ``None``, or for either
            ``requests`` or ``sensors`` to be absent from `model_changes`.
        """
        self._state_cb = cb

    def close(self):
        self.stop()

    def start(self, timeout=None):
        """
        Note: always call stop() and wait until_stopped() when you are done
        with the container to make sure the container cleans up correctly.
        """
        return self.connect(timeout)

    def stop(self, timeout=None):
        self._running = False
        self.katcp_client.stop(timeout)

    def until_stopped(self, timeout=None):
        """Return future that resolves when the client has stopped

        See the `DeviceClient.until_stopped` docstring for parameter
        definitions and more info.
        """
        return self.katcp_client.until_stopped(timeout)

    def join(self, timeout=None):
        self.katcp_client.join(timeout)

    def _update_index(self, index, name, data):
        if name not in index:
            index[name] = data
        else:
            orig_data = index[name]
            for key, value in data.items():
                if orig_data.get(key) != value:
                    orig_data[key] = value
                    orig_data['_changed'] = True

    def handle_sensor_value(self):
        """Handle #sensor-value informs just like #sensor-status informs"""
        self.katcp_client.hook_inform('sensor-value',
                                      self._cb_inform_sensor_status)

    @tornado.gen.coroutine
    def inspect(self):
        """Inspect device requests and sensors, update model.

        Returns
        -------

        Tornado future that resolves with:

        model_changes : Nested AttrDict or None
            Contains sets of added/removed request/sensor names

            Example structure:
            ::

                {
                    'requests':{
                        'added': set(['req1', 'req2']),
                        'removed': set(['req10', 'req20'])
                    }
                    'sensors': {
                        'added': set(['sens1', 'sens2']),
                        'removed': set(['sens10', 'sens20'])
                    }
                }

            If there are no changes keys may be omitted. If an item is in both
            the 'added' and 'removed' sets that means that it changed.

            If neither request not sensor changes are present, None is returned
            instead of a nested structure.

        """
        timeout_manager = future_timeout_manager(self.sync_timeout)
        sensor_index_before = copy.copy(self._sensors_index)
        request_index_before = copy.copy(self._requests_index)
        try:
            request_changes = yield self.inspect_requests(
                timeout=timeout_manager.remaining())
            sensor_changes = yield self.inspect_sensors(
                timeout=timeout_manager.remaining())
        except Exception:
            # Ensure atomicity of sensor and request updates ; if the one
            # fails, the other should act as if it has failed too.
            self._sensors_index = sensor_index_before
            self._requests_index = request_index_before
            raise

        model_changes = AttrDict()
        if request_changes:
            model_changes.requests = request_changes
        if sensor_changes:
            model_changes.sensors = sensor_changes
        if model_changes:
            raise Return(model_changes)

    @tornado.gen.coroutine
    def inspect_requests(self, name=None, timeout=None):
        """Inspect all or one requests on the device. Update requests index.

        Parameters
        ----------
        name : str or None, optional
            Name of the request or None to get all requests.
        timeout : float or None, optional
            Timeout for request inspection, None for no timeout

        Returns
        -------
        Tornado future that resolves with:

        changes : :class:`~katcp.core.AttrDict`
            AttrDict with keys ``added`` and ``removed`` (of type
            :class:`set`), listing the requests that have been added or removed
            respectively.  Modified requests are listed in both. If there are
            no changes, returns ``None`` instead.

            Example structure:
            ::

                {
                    'added': set(['req1', 'req2']),
                    'removed': set(['req10', 'req20'])
                }

        """
        maybe_timeout = future_timeout_manager(timeout)
        if name is None:
            msg = katcp.Message.request('help')
        else:
            msg = katcp.Message.request('help', name)
        reply, informs = yield self.katcp_client.future_request(
            msg, timeout=maybe_timeout.remaining())

        if not reply.reply_ok():
            # If an unknown request is specified the desired result is to return
            # an empty list even though the request will fail
            if name is None or b'Unknown request' not in reply.arguments[1]:
                raise SyncError(
                    'Error reply during sync process for {}: {}'
                    .format(self.bind_address_string, reply))
        # Get recommended timeouts hints for slow requests if the server
        # provides them
        timeout_hints_available = (
            self.katcp_client.protocol_flags.request_timeout_hints)
        if timeout_hints_available:
            timeout_hints = yield self._get_request_timeout_hints(
                name, timeout=maybe_timeout.remaining())
        else:
            timeout_hints = {}

        requests_old = set(self._requests_index.keys())
        requests_updated = set()
        for msg in informs:
            req_name = ensure_native_str(msg.arguments[0])
            req = {'name': req_name,
                   'description': ensure_native_str(msg.arguments[1]),
                   'timeout_hint': timeout_hints.get(req_name)}
            requests_updated.add(req_name)
            self._update_index(self._requests_index, req_name, req)

        added, removed = self._difference(
            requests_old, requests_updated, name, self._requests_index)
        if added or removed:
            raise Return(AttrDict(added=added, removed=removed))

    @tornado.gen.coroutine
    def _get_request_timeout_hints(self, name=None, timeout=None):
        """Get request timeout hints from device

        Parameters
        =========
        name : str or None, optional
            Name of the request or None to get all request timeout hints.
        timeout : float seconds
            Timeout for ?request-timeout-hint

        Returns
        -------
        Tornado future that resolves with:

        timeout_hints : dict request_name -> timeout_hint

        where

        request_name : str
            Name of the request
        timeout_hint : float
            Suggested request timeout hint from device ?request-timeout_hint

        Note, if there is no request hint, there will be no entry in the
        dict. If you request the hint for a named request that has no hint, an
        empty dict will be returned.

        """
        timeout_hints = {}
        req_msg_args = ['request-timeout-hint']
        if name:
            req_msg_args.append(name)
        req_msg = katcp.Message.request(*req_msg_args)
        reply, informs = yield self.katcp_client.future_request(
            req_msg, timeout=timeout)
        if not reply.reply_ok():
            raise SyncError('Error retrieving request timeout hints: "{}"\n'
                            'in reply to request {}, continuing with sync'
                            .format(reply, req_msg))
        for inform in informs:
            request_name = ensure_native_str(inform.arguments[0])
            timeout_hint = float(inform.arguments[1])
            if timeout_hint > 0:
                timeout_hints[request_name] = timeout_hint
        raise Return(timeout_hints)

    @tornado.gen.coroutine
    def inspect_sensors(self, name=None, timeout=None):
        """Inspect all or one sensor on the device. Update sensors index.

        Parameters
        ----------
        name : str or None, optional
            Name of the sensor or None to get all sensors.
        timeout : float or None, optional
            Timeout for sensors inspection, None for no timeout

        Returns
        -------
        Tornado future that resolves with:

        changes : :class:`~katcp.core.AttrDict`
            AttrDict with keys ``added`` and ``removed`` (of type
            :class:`set`), listing the sensors that have been added or removed
            respectively.  Modified sensors are listed in both. If there are no
            changes, returns ``None`` instead.

            Example structure:
            ::

                {
                    'added': set(['sens1', 'sens2']),
                    'removed': set(['sens10', 'sens20'])
                }

        """
        if name is None:
            msg = katcp.Message.request('sensor-list')
        else:
            msg = katcp.Message.request('sensor-list', name)

        reply, informs = yield self.katcp_client.future_request(
            msg, timeout=timeout)
        self._logger.debug('{} received {} sensor-list informs, reply: {}'
            .format(self.bind_address_string, len(informs), reply))
        if not reply.reply_ok():
            # If an unknown sensor is specified the desired result is to return
            # an empty list, even though the request will fail
            if name is None or b'Unknown sensor' not in reply.arguments[1]:
                raise SyncError('Error reply during sync process: {}'
                                .format(reply))


        sensors_old = set(self._sensors_index.keys())
        sensors_updated = set()
        for msg in informs:
            sen_name = ensure_native_str(msg.arguments[0])
            sensors_updated.add(sen_name)
            sen = {'description': ensure_native_str(msg.arguments[1]),
                   'units': ensure_native_str(msg.arguments[2]),
                   'sensor_type': ensure_native_str(msg.arguments[3]),
                   'params': [ensure_native_str(arg) for arg in msg.arguments[4:]]}
            self._update_index(self._sensors_index, sen_name, sen)

        added, removed = self._difference(
            sensors_old, sensors_updated, name, self._sensors_index)

        for sensor_name in removed:
            if sensor_name in self._sensor_object_cache:
                del self._sensor_object_cache[sensor_name]

        if added or removed:
            raise Return(AttrDict(added=added, removed=removed))

    @tornado.gen.coroutine
    def future_check_sensor(self, name, update=None):
        """Check if the sensor exists.

        Used internally by future_get_sensor. This method is aware of
        synchronisation in progress and if inspection of the server is allowed.

        Parameters
        ----------
        name : str
            Name of the sensor to verify.
        update : bool or None, optional
            If a katcp request to the server should be made to check if the
            sensor is on the server now.

        Notes
        -----
        Ensure that self.state.data_synced == True if yielding to
        future_check_sensor from a state-change callback, or a deadlock will
        occur.

        """
        exist = False
        yield self.until_data_synced()
        if name in self._sensors_index:
            exist = True
        else:
            if update or (update is None and self._update_on_lookup):
                yield self.inspect_sensors(name)
                exist = yield self.future_check_sensor(name, False)

        raise tornado.gen.Return(exist)

    @tornado.gen.coroutine
    def future_get_sensor(self, name, update=None):
        """Get the sensor object.

        Check if we have information for this sensor, if not connect to server
        and update (if allowed) to get information.

        Parameters
        ----------
        name : string
            Name of the sensor.
        update : bool or None, optional
            True allow inspect client to inspect katcp server if the sensor
            is not known.

        Returns
        -------
        Sensor created by :meth:`sensor_factory` or None if sensor not found.

        Notes
        -----
        Ensure that self.state.data_synced == True if yielding to future_get_sensor from
        a state-change callback, or a deadlock will occur.

        """
        obj = None
        exist = yield self.future_check_sensor(name, update)
        if exist:
            sensor_info = self._sensors_index[name]
            obj = sensor_info.get('obj')
            if obj is None:
                params = []
                for param in sensor_info.get('params', []):
                    params.append(native_str_to_bytes(param))
                sensor_type = katcp.Sensor.parse_type(
                    sensor_info.get('sensor_type'))
                sensor_params = katcp.Sensor.parse_params(
                    sensor_type,
                    params)
                obj = self.sensor_factory(
                    name=name,
                    sensor_type=sensor_type,
                    description=sensor_info.get('description'),
                    units=sensor_info.get('units'),
                    params=sensor_params)
                self._sensors_index[name]['obj'] = obj
                self._sensor_object_cache[name] = obj

        raise tornado.gen.Return(obj)

    @tornado.gen.coroutine
    def future_check_request(self, name, update=None):
        """Check if the request exists.

        Used internally by future_get_request. This method is aware of
        synchronisation in progress and if inspection of the server is allowed.

        Parameters
        ----------
        name : str
            Name of the request to verify.
        update : bool or None, optional
            If a katcp request to the server should be made to check if the
            sensor is on the server. True = Allow, False do not Allow, None
            use the class default.

        Notes
        -----
        Ensure that self.state.data_synced == True if yielding to future_check_request
        from a state-change callback, or a deadlock will occur.

        """
        exist = False
        yield self.until_data_synced()
        if name in self._requests_index:
            exist = True
        else:
            if update or (update is None and self._update_on_lookup):
                yield self.inspect_requests(name)
                exist = yield self.future_check_request(name, False)
        raise tornado.gen.Return(exist)

    @tornado.gen.coroutine
    def future_get_request(self, name, update=None):
        """Get the request object.

        Check if we have information for this request, if not connect to server
        and update (if allowed).

        Parameters
        ----------
        name : string
            Name of the request.
        update : bool or None, optional
            True allow inspect client to inspect katcp server if the request
            is not known.

        Returns
        -------
        Request created by :meth:`request_factory` or None if request not found.

        Notes
        -----
        Ensure that self.state.data_synced == True if yielding to future_get_request
        from a state-change callback, or a deadlock will occur.

        """
        obj = None
        exist = yield self.future_check_request(name, update)
        if exist:
            request_info = self._requests_index[name]
            obj = request_info.get('obj')
            if obj is None:
                obj = self.request_factory(**request_info)
                self._requests_index[name]['obj'] = obj

        raise tornado.gen.Return(obj)

    @tornado.gen.coroutine
    def update_sensor(self, name, timestamp, status, value):
        sensor = self._sensor_object_cache.get(name)
        # TODO (NM) Performance improvement idea: Make this a normal function
        # that only handles the cached sensor object case (should be the common
        # case). If cached sensor is not found, kick off a coroutine that
        # handles the future_get_sensor case.
        if not sensor:
            sensor = yield self.future_get_sensor(name)
        if sensor:
            # TODO (NM) Performance idea: probably we can cache katcp_major at
            # sync time?
            katcp_major = self.katcp_client.protocol_flags.major
            sensor.set_formatted(timestamp, status, value, katcp_major)
        else:
            self._logger.error('Received update for "%s", but could not create'
                               ' sensor object.' % name)

    def _cb_connection_state(self, connected):
        if connected:
            self._disconnected.clear()
            self._connected.set()
        else:
            self._connected.clear()
            self._disconnected.set()

    def _cb_inform_sensor_status(self, msg):
        """Update received for an sensor."""
        timestamp = msg.arguments[0]
        num_sensors = int(msg.arguments[1])
        assert len(msg.arguments) == 2 + num_sensors * 3
        for n in range(num_sensors):
            # Note: update_sensor needs byte strings, except for sensor name
            name = ensure_native_str(msg.arguments[2 + n * 3])
            status = msg.arguments[3 + n * 3]
            value = msg.arguments[4 + n * 3]
            self.update_sensor(name, timestamp, status, value)

    def _cb_inform_interface_change(self, msg):
        """Update the sensors and requests available."""
        self._logger.debug('cb_inform_interface_change(%s)', msg)
        self._interface_changed.set()

    def _cb_inform_deprecated(self, msg):
        """Log a message that an deprecated inform has been received.."""
        self._logger.warning("Received a deprecated inform: {0}."
                             .format(msg.name))

    def simple_request(self, request, *args, **kwargs):
        """Create and send a request to the server.

        This method implements a very small subset of the options
        possible to send an request. It is provided as a shortcut to
        sending a simple request.

        Parameters
        ----------
        request : str
            The request to call.
        *args : list of objects
            Arguments to pass on to the request.

        Keyword Arguments
        -----------------
        timeout : float or None, optional
            Timeout after this amount of seconds (keyword argument).
        mid : None or int, optional
            Message identifier to use for the request message. If None, use either
            auto-incrementing value or no mid depending on the KATCP protocol version
            (mid's were only introduced with KATCP v5) and the value of the `use_mid`
            argument. Defaults to None
        use_mid : bool
            Use a mid for the request if True. Defaults to True if the server supports
            them.

        Returns
        -------
        future object.

        Examples
        --------
        ::

            reply, informs = yield ic.simple_request('help', 'sensor-list')

        """
        # TODO (NM 2016-11-03) This method should really live on the lower level
        # katcp_client in client.py, is generally useful IMHO
        use_mid = kwargs.get('use_mid')
        timeout = kwargs.get('timeout')
        mid = kwargs.get('mid')
        msg = katcp.Message.request(request, *args, mid=mid)
        return self.katcp_client.future_request(msg, timeout, use_mid)

    def _difference(self, original_keys, updated_keys, name, item_index):
        """Calculate difference between the original and updated sets of keys.

        Removed items will be removed from item_index, new items should have
        been added by the discovery process. (?help or ?sensor-list)

        This method is for use in inspect_requests and inspect_sensors only.

        Returns
        -------

        (added, removed)
        added : set of str
            Names of the keys that were added
        removed : set of str
            Names of the keys that were removed

        """
        original_keys = set(original_keys)
        updated_keys = set(updated_keys)
        added_keys = updated_keys.difference(original_keys)
        removed_keys = set()
        if name is None:
            removed_keys = original_keys.difference(updated_keys)
        elif name not in updated_keys and name in original_keys:
            removed_keys = set([name])

        for key in removed_keys:
            if key in item_index:
                del(item_index[key])

        # Check the keys that was not added now or not lined up for removal,
        # and see if they changed.
        for key in updated_keys.difference(added_keys.union(removed_keys)):
            if item_index[key].get('_changed'):
                item_index[key]['_changed'] = False
                removed_keys.add(key)
                added_keys.add(key)

        return added_keys, removed_keys



# Copyright 2016 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import collections
import gc
import logging
import time
import unittest
import weakref

from builtins import object, range

import mock
import tornado

import katcp

from katcp import Message, Sensor, inspecting_client
from katcp.inspecting_client import InspectingClientAsync
from katcp.testutils import (DeviceTestSensor, DeviceTestServer,
                             DeviceTestServerWithTimeoutHints,
                             start_thread_with_cleanup)

logger = logging.getLogger(__name__)

class TestExponentialRandomBackoff(unittest.TestCase):
    def test_delays(self):
        with mock.patch(inspecting_client.__name__+'.random.random') as m_rand:
            # Parameters calculated so that 4 calls to fail will be < delay_max,
            rv = m_rand.return_value = 0.4
            di = 1.5
            dm = 12
            ex = 2
            r = 0.7
            kwargs = dict(
                delay_initial=di, delay_max=dm, exp_fac=ex, randomicity=r)
            DUT = inspecting_client.ExponentialRandomBackoff(**kwargs)
            first_delay = DUT.delay
            self.assertAlmostEqual(first_delay, (1-r)*di + r*di*rv)
            for i in range(1,4):
                DUT.failed()
                self.assertAlmostEqual(DUT.delay, first_delay*(ex**i))

            # Do several more failures
            [DUT.failed() for x in range(5)]
            # Check that the delay does not grow beyond the maximum expected
            # value
            max_expected_delay = (1-r)*dm + r*dm*rv
            self.assertAlmostEqual(DUT.delay, max_expected_delay)

            # Test that DUT.success() resets the delay
            DUT.success()
            self.assertEqual(DUT.delay, first_delay)


class TestICAClass(tornado.testing.AsyncTestCase):

    def setUp(self):
        super(TestICAClass, self).setUp()
        self.client = InspectingClientAsync('', 0, initial_inspection=False,
                                            ioloop=self.io_loop)

        self.client.set_state_callback(self._cb_state)

    def test_initial_state(self):
        self.assertEqual(self.client.state,
                         inspecting_client.InspectingClientStateType(
            connected=False, synced=False, model_changed=False, data_synced=False))

    def _cb_state(self, state, model_changes):
        """A callback used in the test."""
        self.stop((state, model_changes))

    def test_util_method_difference_add(self):
        """Test the _difference utility method on add."""

        name = None
        original_keys = ['B', 'C']
        updated_keys = ['B', 'C', 'D']
        self.client._sensors_index = {}
        for sen in original_keys:
            data = {'description': "This is {0}.".format(sen)}
            self.client._update_index(self.client._sensors_index, sen, data)
        added, removed = self.client._difference(
            original_keys, updated_keys, name, self.client._sensors_index)
        self.assertIn('D', added)
        self.assertIn('B', self.client.sensors)
        self.assertIn('C', self.client.sensors)

    def test_util_method_difference_rem(self):
        """Test the _difference utility method on remove."""

        name = None
        original_keys = ['A', 'B', 'C']
        updated_keys = ['B', 'C']
        self.client._sensors_index = {}
        for sen in original_keys:
            data = {'description': "This is {0}.".format(sen)}
            self.client._update_index(self.client._sensors_index, sen, data)

        added, removed = self.client._difference(
            original_keys, updated_keys, name, self.client._sensors_index)
        self.assertIn('A', removed)
        self.assertNotIn('A', self.client.sensors)
        self.assertIn('B', self.client.sensors)
        self.assertIn('C', self.client.sensors)

    def test_util_method_difference_rem_named(self):
        """Test the _difference utility method on remove with name set."""

        name = 'A'
        original_keys = ['B', 'C']
        updated_keys = []
        self.client._sensors_index = {}
        for sen in original_keys:
            data = {'description': "This is {0}.".format(sen)}
            self.client._update_index(self.client._sensors_index, sen, data)
        added, removed = self.client._difference(
            original_keys, updated_keys, name, self.client._sensors_index)
        self.assertNotIn('A', removed)
        self.assertNotIn('A', self.client.sensors)
        self.assertIn('B', self.client.sensors)
        self.assertIn('C', self.client.sensors)

    def test_util_method_difference_changed(self):
        """Test the _difference utility method on changed."""

        name = None
        original_keys = ['B']
        updated_keys = ['B']

        self.client._sensors_index = {}
        for sen in original_keys:
            data = {'description': "This is {0}.".format(sen), '_changed': True}
            self.client._update_index(self.client._sensors_index, sen, data)

        added, removed = self.client._difference(original_keys, updated_keys,
                                                 name, self.client._sensors_index)
        self.assertIn('B', self.client.sensors)

    def test_update_index(self):
        """Test the update_index method."""

        index = self.client._sensors_index = {}

        data = {'description': "This is {0}.".format('A')}
        self.client._update_index(index, 'A', data)

        data = {'description': "This is {0}.".format('B')}
        self.client._update_index(index, 'B', data)

        data = {'description': "This is {0}.".format('A')}
        self.client._update_index(index, 'A', data)

        data = {'description': "This is new {0}.".format('B')}
        self.client._update_index(index, 'B', data)

        self.assertIn('new', index['B'].get('description'))
        self.assertFalse(index['A'].get('_changed', False))
        self.assertTrue(index['B'].get('_changed'))


class TestInspectingClientInspect(tornado.testing.AsyncTestCase):
    """Check that inspection populates the request/sensor index correctly"""

    def _get_server(self, hints):
        """Return a running test server with or without request timeout hints

        Parameters
        ----------
        hints : bool
            Whether or not the server should have request timeout hints enabled

        """
        ServerClass = (DeviceTestServerWithTimeoutHints if hints
                       else DeviceTestServer)
        server = ServerClass('', 0)
        start_thread_with_cleanup(self, server, start_timeout=1)
        host, port = server.bind_address
        return host, port, server

    def _get_expected_request_index(self, server):
        """Return expected request index data structure for given server instance

        Parameters
        ----------
        server : :class:`DeviceServer` instance

        """
        hints = getattr(server, 'request_timeout_hints', {})
        expected = {}
        for req, handler in list(server._request_handlers.items()):
            expected[req] = {'name': req,
                             'description': handler.__doc__,
                             'timeout_hint': hints.get(req)}
        return expected

    @tornado.gen.coroutine
    def _test_inspect_requests(self, timeout_hints):
        """Test  index creation

        Parameters
        ----------
        timeout_hints : bool
            Whether or not the server being tested against should privide
            request timeout hints

        Disables :meth:`InspectingClient.inspect_requests` so that inspecting
        does not happen until the test triggers is. Then checks that
        :attr:`InspectingClient._requests_index` is correctly generated

        """
        host, port, server = self._get_server(hints=timeout_hints)
        DUT = InspectingClientAsync(host, port, ioloop=self.io_loop)
        # mock out the state loop so that syncing does not happen automatically
        DUT._state_loop = mock.Mock()
        # Connect to test server
        yield DUT.connect(timeout=1)
        # Now run the method under test
        yield DUT.inspect_requests()
        expected_request_index = self._get_expected_request_index(server)
        self.assertEqual(DUT._requests_index, expected_request_index)

    @tornado.testing.gen_test
    def test_inspect_request_no_timeout_hints(self):
        yield self._test_inspect_requests(timeout_hints=False)

    @tornado.testing.gen_test
    def test_inspect_request_with_timeout_hints(self):
        yield self._test_inspect_requests(timeout_hints=True)

    # TODO NM 2017-04-12 Tests should be added for sensor index. I just added
    # the minimum needed to test new functionality for CB-569


class TestInspectingClientAsync(tornado.testing.AsyncTestCase):

    def setUp(self):
        super(TestInspectingClientAsync, self).setUp()
        self.server = DeviceTestServer('', 0)
        start_thread_with_cleanup(self, self.server, start_timeout=1)
        self.host, self.port = self.server.bind_address

        self.client = InspectingClientAsync(self.host, self.port,
                                            ioloop=self.io_loop)
        self.io_loop.add_callback(self.client.connect)

    @tornado.testing.gen_test
    def test_timeout_of_until_synced(self):
        # Test for timing out
        with self.assertRaises(tornado.gen.TimeoutError):
            yield self.client.until_synced(timeout=0.00001)
        # Test for NOT timing out
        yield self.client.until_synced(timeout=0.5)

    @tornado.testing.gen_test
    def test_simple_request(self):
        """Perform a basic request."""
        yield self.client.until_synced()
        reply, informs = yield self.client.simple_request('help', 'watchdog')
        self.assertIn('ok', str(reply))
        self.assertEqual(len(informs), 1)

    @tornado.testing.gen_test
    def test_sensor(self):
        """Access the sensor with the Async client."""
        yield self.client.until_synced()
        sensor_name = 'an.int'
        sensor = yield self.client.future_get_sensor(sensor_name)
        self.assertEqual(sensor.name, sensor_name)
        self.assertEqual(sensor.stype, 'integer')

        sensor_name = 'a.discrete'
        sensor = yield self.client.future_get_sensor(sensor_name)
        self.assertEqual(sensor.name, sensor_name)
        self.assertEqual(sensor.stype, 'discrete')

        sensor_name = 'a.float'
        sensor = yield self.client.future_get_sensor(sensor_name)
        self.assertEqual(sensor.name, sensor_name)
        self.assertEqual(sensor.stype, 'float')

        # Unknown sensor requests return a None.
        sensor_name = 'thing.unknown_sensor'
        sensor = yield self.client.future_get_sensor(sensor_name)
        self.assertIsNone(sensor)

    @tornado.testing.gen_test
    def test_request_access(self):
        """Test access to requests."""
        yield self.client.until_synced()

        request_name = 'watchdog'
        self.assertIn(request_name, self.client.requests)
        request = yield self.client.future_get_request(request_name)
        self.assertEqual(request.name, request_name)
        self.assertTrue(request.description,
                        'Expected an description: got nothing.')

        # Unknown request return a None.
        request_name = 'watchcat'
        self.assertNotIn(request_name, self.client.requests)
        request = yield self.client.future_get_request(request_name)
        self.assertIsNone(request)

    @tornado.testing.gen_test
    def test_sensor_add_remove(self):
        """Test a sensor being added and then remove it."""
        yield self.client.until_synced()

        sensor = DeviceTestSensor(Sensor.INTEGER, "another.int",
                                  "An Integer.",
                                  "count", [-5, 5], timestamp=time.time(),
                                  status=Sensor.NOMINAL, value=3)
        # Check that the sensor does not exist currently
        self.assertNotIn(sensor.name, self.client.sensors)

        # Add a sensor.
        self.server.add_sensor(sensor)
        self.server.mass_inform(Message.inform('interface-changed'))
        # Do a blocking request to ensure #interface-changed has been received
        yield self.client.simple_request('watchdog')
        yield self.client.until_synced()
        self.assertIn('another.int', self.client.sensors)

        # Remove a sensor.
        self.server.remove_sensor(sensor)
        self.server.mass_inform(Message.inform('interface-changed'))
        # Do a blocking request to ensure #interface-changed has been received
        yield self.client.simple_request('watchdog')

        yield self.client.until_synced()
        self.assertNotIn('another.int', self.client.sensors)


    @tornado.testing.gen_test
    def test_request_add_remove(self):
        """Test a request being added and then remove it."""
        yield self.client.until_synced()

        def request_sparkling_new(self, req, msg):
            """A new command."""
            return Message.reply(msg.name, "ok", "bling1", "bling2")

        # Check that the request did not exist before
        self.assertNotIn('sparkling-new', self.client.requests)

        # Add a request.
        self.server.request_sparkling_new = request_sparkling_new
        self.server._request_handlers['sparkling-new'] = request_sparkling_new
        self.server.mass_inform(Message.inform('interface-changed'))
        # Do a blocking request to ensure #interface-changed has been received
        yield self.client.simple_request('watchdog')

        yield self.client.until_synced()
        self.assertIn('sparkling-new', self.client.requests)
        req = yield self.client.future_get_request('sparkling-new')
        self.assertEqual(req.name, 'sparkling-new')

        # Remove a request.
        self.server.request_sparkling_new = None
        del(self.server._request_handlers['sparkling-new'])
        self.server.mass_inform(Message.inform('interface-changed'))
        # Do a blocking request to ensure #interface-changed has been received
        self.client.simple_request('watchdog')
        yield self.client.until_synced()
        self.assertNotIn('sparkling_new', self.client.requests)

    @tornado.testing.gen_test
    def test_send_request(self):
        """Very high level test.

        Calling methods to insure they do not raise exception.

        """
        client = InspectingClientAsync(self.host, self.port,
                                       ioloop=self.io_loop,
                                       initial_inspection=False)

        yield client.connect()
        yield client.until_connected()
        yield client.until_synced()
        self.assertEqual(len(client.sensors), 0)

        self.assertEqual(len(client.requests), 0)
        self.assertTrue(client.synced)
        self.assertTrue(client.is_connected())
        self.assertTrue(client.connected)
        yield client.simple_request('sensor-sampling', 'an.int', 'event')
        # Wait for sync and check if the sensor was automaticaly added.
        # Get the sensor object and see if it has data.
        sensor = yield client.future_get_sensor('an.int')
        self.assertEqual(len(client.sensors), 1)
        self.assertTrue(sensor.read())
        self.assertEqual(len(client.requests), 0)

    @tornado.testing.gen_test
    def test_handle_sensor_value(self):
        yield self.client.until_connected()
        # Test that #sensor-value informs are handles like #sensor-inform informs if
        # handle_sensor_value() is called.
        sens = yield self.client.future_get_sensor('an.int')
        test_val = 1911
        # Check that the sensor doesn't start out with our test value
        self.assertNotEqual(sens.read().value, test_val)
        # Now set the value of the sensor on the server
        server_sens = self.server.get_sensor('an.int')
        server_sens.set_value(test_val)
        # Do a request to ensure we are synced with the server
        yield self.client.simple_request('watchdog')
        # Test that the value has not yet been set on the client sensor (there are no
        # strategies set)
        self.assertNotEqual(sens.read().value, test_val)
        # Now do a ?sensor-value request, and check that our object is NOT YET updated,
        # since we have not called handle_sensor_value() yet
        yield self.client.simple_request('sensor-value', 'an.int')
        self.assertNotEqual(sens.read().value, test_val)

        # Test call
        self.client.handle_sensor_value()

        # Now do a ?sensor-value request, and check that our object is updated
        yield self.client.simple_request('sensor-value', 'an.int')
        self.assertEqual(sens.read().value, test_val)

    @tornado.testing.gen_test
    def test_factories(self):
        yield self.client.until_connected()
        # Test that the correct factories are used to construct sensor and request
        # objects, and that the factories are called with the correct parameters.
        sf = self.client.sensor_factory = mock.Mock()
        rf = self.client.request_factory = mock.Mock()

        sen = yield self.client.future_get_sensor('an.int')
        req = yield self.client.future_get_request('watchdog')
        self.assertIs(sen, sf.return_value)
        sf.assert_called_once_with(
            units='count', sensor_type=0, params=[-5, 5],
            description='An Integer.', name='an.int')
        self.assertIs(req, rf.return_value)
        rf.assert_called_once_with(
            name='watchdog', description=mock.ANY, timeout_hint=None)

    @tornado.testing.gen_test
    def test_no_memory_leak_after_usage(self):
        client = InspectingClientAsync(self.host, self.port, ioloop=self.io_loop)
        wr = weakref.ref(client)

        yield client.connect()
        yield client.until_synced()
        client.stop()
        yield client.until_stopped()
        client.join()

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())


class TestInspectingClientAsyncStateCallback(tornado.testing.AsyncTestCase):
    longMessage = True
    maxDiff = None

    def setUp(self):
        super(TestInspectingClientAsyncStateCallback, self).setUp()
        self.server = DeviceTestServer('', 0)
        start_thread_with_cleanup(self, self.server, start_timeout=1)
        self.host, self.port = self.server.bind_address
        self.state_cb_future = tornado.concurrent.Future()
        self.client = InspectingClientAsync(self.host, self.port,
                                            ioloop=self.io_loop)
        # Set a short initial_resync timeout to make resync tests quick
        self.client.initial_resync_timeout = 0.001
        self.client.set_state_callback(self._test_state_cb)
        self.done_state_cb_futures = []
        self.cnt_state_cb_futures = collections.defaultdict(tornado.concurrent.Future)

    def _test_state_cb(self, state, model_changes):
        f = self.state_cb_future
        self.state_cb_future = tornado.concurrent.Future()
        self.done_state_cb_futures.append(f)
        num_calls = len(self.done_state_cb_futures)
        f.set_result((state, model_changes))
        self.cnt_state_cb_futures[num_calls].set_result(None)

    @tornado.gen.coroutine
    def _check_cb_count(self, expected_count):
        """Let the ioloop run and assert that the callback has been called
        the expected number of times"""
        yield tornado.gen.moment
        self.assertEqual(len(self.done_state_cb_futures), expected_count)

    @tornado.testing.gen_test(timeout=1)
    def test_from_connect(self):
        # Hold back #version-connect informs
        num_calls_before = len(self.done_state_cb_futures)
        logger.debug('before starting client, num_calls_before:{}'.format(num_calls_before))

        self.server.proceed_on_client_connect.clear()
        self.client.connect()

        state, model_changes = yield self.state_cb_future
        self.assertEqual(state, inspecting_client.InspectingClientStateType(
            connected=False, synced=False, model_changed=False, data_synced=False))

        state, model_changes = yield self.state_cb_future
        self.assertEqual(state, inspecting_client.InspectingClientStateType(
            connected=True, synced=False, model_changed=False, data_synced=False))
        self.assertIs(model_changes, None)
        # Due to the structure of the state loop the initial state may be sent
        # twice before we get here, and was the case for the initial
        # implementation. Changes made on 2015-01-26 caused it to happen only
        # once, hence + 1. If the implementation changes having + 2 would also
        # be OK. As, indeed, changes made on 2016-11-03 caused again :)
        yield self._check_cb_count(num_calls_before + 2)

        # Now let the server send #version-connect informs
        num_calls_before = len(self.done_state_cb_futures)
        self.server.ioloop.add_callback(self.server.proceed_on_client_connect.set)
        # We're expecting two calls hard on each other's heels, so lets wait for them
        yield self.cnt_state_cb_futures[num_calls_before + 2]
        # We expected two status callbacks, and no more after
        yield self._check_cb_count(num_calls_before + 2)
        state, model_changes = yield self.done_state_cb_futures[-2]
        state2, model_changes2 = yield self.done_state_cb_futures[-1]
        self.assertEqual(state, inspecting_client.InspectingClientStateType(
            connected=True, synced=False, model_changed=True, data_synced=True))
        # Check that the expected model changes came from the callback
        self._test_expected_model_changes(model_changes)
        self.assertEqual(state2, inspecting_client.InspectingClientStateType(
            connected=True, synced=True, model_changed=False, data_synced=True))
        self.assertEqual(model_changes2, None)

    def _test_expected_model_changes(self, model_changes):
        # Check that the model_changes reflect the sensors and requests of the
        # test sever (self.server)
        server_sensors = list(self.server._sensors.keys())
        server_requests = list(self.server._request_handlers.keys())
        self.assertEqual(model_changes, dict(
            sensors=dict(added=set(server_sensors), removed=set()),
            requests=dict(added=set(server_requests), removed=set())))

    @tornado.testing.gen_test(timeout=1)
    def test_reconnect(self):
        self.client.connect()
        yield self.client.until_synced()
        yield tornado.gen.moment   # Make sure the ioloop is 'caught up'

        # cause a disconnection and check that the callback is called
        num_calls_before = len(self.done_state_cb_futures)
        self.server.stop()
        self.server.join()
        state, model_changes = yield self.state_cb_future
        self.assertEqual(state, inspecting_client.InspectingClientStateType(
            connected=False, synced=False, model_changed=False, data_synced=False))
        self.assertIs(model_changes, None)
        yield self._check_cb_count(num_calls_before + 1)

    @tornado.testing.gen_test(timeout=1)
    def test_stop(self):
        self.client.connect()
        yield self.client.until_synced()
        yield tornado.gen.moment   # Make sure the ioloop is 'caught up'

        # stop and check that the callback is called
        num_calls_before = len(self.done_state_cb_futures)
        next_state_cb_future = self.state_cb_future
        self.client.stop()
        yield self.client.until_stopped()
        self.client.join()
        state, model_changes = yield next_state_cb_future
        self.assertEqual(state, inspecting_client.InspectingClientStateType(
            connected=False, synced=False, model_changed=False, data_synced=False))
        self.assertIs(model_changes, None)
        yield self._check_cb_count(num_calls_before + 1)

    @tornado.gen.coroutine
    def _test_inspection_error(self, break_var, break_message):
        # Test that the client retries if there is an error in the inspection
        # process
        setattr(self.server, break_var, break_message)

        self.client.connect()
        # Wait for the client to be connected
        yield self.client.until_connected()
        # Wait for the state loop to send another update or 2
        yield self.state_cb_future
        state, _ = yield self.state_cb_future
        # Check that data is still not synced
        self.assertFalse(state.synced)
        self.assertFalse(state.data_synced)

        # Now fix the inspection request, client should sync up.
        setattr(self.server, break_var, False)
        # Check that the server's sensors and request are reflected in the model
        # changes.
        #
        changes_state = inspecting_client.InspectingClientStateType(
            connected=True, synced=False, model_changed=True, data_synced=True)
        yield self.client.until_state(changes_state)
        state, model_changes = self.done_state_cb_futures[-1].result()
        assert state == changes_state
        self._test_expected_model_changes(model_changes)
        yield self.client.until_synced()
        self.assertTrue(self.client.synced)

    @tornado.testing.gen_test(timeout=1)
    def test_help_inspection_error(self):
        yield self._test_inspection_error('break_help', 'Help is broken')

    @tornado.testing.gen_test(timeout=1)
    def test_sensor_list_inspection_error(self):
        yield self._test_inspection_error(
            'break_sensor_list', 'Sensor-list is broken')

class Test_InformHookDeviceClient(tornado.testing.AsyncTestCase):
    def setUp(self):
        super(Test_InformHookDeviceClient, self).setUp()
        self.server = DeviceTestServer('', 0)
        start_thread_with_cleanup(self, self.server, start_timeout=1)
        self.host, self.port = self.server.bind_address
        self.client = katcp.inspecting_client._InformHookDeviceClient(
            self.host, self.port)
        self.client.set_ioloop(self.io_loop)
        self.io_loop.add_callback(self.client.start)

    @tornado.testing.gen_test()
    def test_hook_inform(self):
        h1_calls, h2_calls = [], []
        h1 = lambda msg: h1_calls.append(msg)
        # Specifically test that method callback functions are handled properly. See
        # http://stackoverflow.com/questions/13348031/python-bound-and-unbound-method-object
        class MH(object):
            def h2(self, msg):
                h2_calls.append(msg)
        mh = MH()
        # Add same hook multiple times to check that duplicates are removed
        self.client.hook_inform('help', h1)
        self.client.hook_inform('help', h1)
        self.client.hook_inform('sensor-list', mh.h2)
        self.client.hook_inform('sensor-list', mh.h2)

        yield self.client.until_protocol()
        yield self.client.future_request(katcp.Message.request('help', 'watchdog'))
        yield self.client.future_request(katcp.Message.request('sensor-list', 'an.int'))

        self.assertEqual(len(h1_calls), 1)
        self.assertEqual(h1_calls[0].name, 'help')
        self.assertEqual(len(h2_calls), 1)
        self.assertEqual(h2_calls[0].name, 'sensor-list')



# test_kattypes.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Tests for the kattypes module.
   """

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import unittest

from builtins import object

import future
import mock

from katcp import AsyncReply, FailReply, Message
from katcp.core import DEFAULT_KATCP_MAJOR
from katcp.kattypes import (Address, Bool, Discrete, DiscreteMulti, Float, Int,
                            Lru, Regex, Str, StrictTimestamp, Struct,
                            Timestamp, TimestampOrNow, inform, pack_types,
                            request, return_reply, send_reply, unpack_types)

MS_TO_SEC_FAC = 1/1000.
SEC_TO_MS_FAC = 1000


class TestType(unittest.TestCase):
    def setUp(self):
        self._pack = []
        self._unpack = []

    def test_pack(self):
        for t, value, result in self._pack:
            if type(result) is type and issubclass(result, Exception):
                self.assertRaises(result, t.pack, value)
            else:
                packed = t.pack(value)
                self.assertEqual(packed, result)
                self.assertEqual(type(packed), type(result))

    def test_unpack(self):
        for t, value, result in self._unpack:
            if type(result) is type and issubclass(result, Exception):
                self.assertRaises(result, t.unpack, value)
            else:
                unpacked = t.unpack(value)
                self.assertEqual(unpacked, result)
                self.assertEqual(type(unpacked), type(result))



class TestInt(TestType):

    def setUp(self):
        basic = Int()
        default = Int(default=11)
        optional = Int(optional=True)
        default_optional = Int(default=11, optional=True)
        minmax = Int(min=5, max=6)
        big_minmax = Int(min=-2**64, max=2**64)

        self._pack = [
            (basic, 5, b"5"),
            (basic, -5, b"-5"),
            (basic, "a", TypeError),
            (basic, None, ValueError),
            (minmax, 5, b"5"),
            (minmax, 6, b"6"),
            (minmax, 4, ValueError),
            (minmax, 7, ValueError),
            (big_minmax, 2**64, b"18446744073709551616"),
            (big_minmax, -2**64, b"-18446744073709551616"),
            (big_minmax, 2**64 + 1, ValueError),
            (big_minmax, -2**64 - 1, ValueError),
            (default, None, b"11"),
            (default_optional, None, b"11"),
            (optional, None, ValueError),
        ]

        self._unpack = [
            (basic, b"5", 5),
            (basic, b"-5", -5),
            (basic, b"a", ValueError),
            (basic, None, ValueError),
            (minmax, b"5", 5),
            (minmax, b"6", 6),
            (minmax, b"4", ValueError),
            (minmax, b"7", ValueError),
            (default, None, 11),
            (default_optional, None, 11),
            (optional, None, None),
        ]


class TestFloat(TestType):

    def setUp(self):
        basic = Float()
        default = Float(default=11.0)
        optional = Float(optional=True)
        default_optional = Float(default=11.0, optional=True)
        minmax = Float(min=5.0, max=6.0)
        big_minmax = Float(min=-2**64, max=2**64)  # allow integer limits

        self._pack = [
            (basic, 5.0, b"5.0"),
            (basic, -5.0, b"-5.0"),
            (basic, 5.5, b"5.5"),
            (basic, -5.5, b"-5.5"),
            (basic, "a", ValueError),
            (basic, None, ValueError),
            (minmax, 5.0, b"5.0"),
            (minmax, 6.0, b"6.0"),
            (minmax, 4.5, ValueError),
            (minmax, 6.5, ValueError),
            (big_minmax, 2**64, b"1.8446744073709552e+19"),
            (big_minmax, -2**64, b"-1.8446744073709552e+19"),
            (big_minmax, 2**64 + 1, ValueError),
            (big_minmax, -2**64 - 1, ValueError),
            (default, None, b"11.0"),
            (default_optional, None, b"11.0"),
            (optional, None, ValueError),
        ]

        self._unpack = [
            (basic, b"5", 5.0),
            (basic, b"5.0", 5.0),
            (basic, b"-5", -5.0),
            (basic, b"-5.0", -5.0),
            (basic, b"5.5", 5.5),
            (basic, b"-5.5", -5.5),
            (basic, b"a", ValueError),
            (basic, None, ValueError),
            (minmax, b"5", 5.0),
            (minmax, b"6", 6.0),
            (minmax, b"4.5", ValueError),
            (minmax, b"6.5", ValueError),
            (default, None, 11.0),
            (default_optional, None, 11.0),
            (optional, None, None),
        ]


class TestBool(TestType):

    def setUp(self):
        basic = Bool()
        default = Bool(default=True)
        optional = Bool(optional=True)
        default_optional = Bool(default=True, optional=True)

        self._pack = [
            (basic, True, b"1"),
            (basic, False, b"0"),
            (basic, 1, b"1"),
            (basic, 0, b"0"),
            (basic, "a", b"1"),
            (basic, None, ValueError),
            (default, None, b"1"),
            (default_optional, None, b"1"),
            (optional, None, ValueError),
        ]

        self._unpack = [
            (basic, b"1", True),
            (basic, b"0", False),
            (basic, b"a", ValueError),
            (basic, None, ValueError),
            (default, None, True),
            (default_optional, None, True),
            (optional, None, None),
        ]


class TestDiscrete(TestType):

    def setUp(self):
        basic = Discrete(("VAL1", "VAL2"))
        default = Discrete(("VAL1", "VAL2"), default="VAL1")
        optional = Discrete(("VAL1", "VAL2"), optional=True)
        default_optional = Discrete(("VAL1", "VAL2"), default="VAL1",
                                    optional=True)
        case_insensitive = Discrete(("val1", "VAL2"), case_insensitive=True)
        values = ('VAL{}'.format(i + 1) for i in range(2))
        basic_generator = Discrete(values)

        self._pack = [
            (basic, "VAL1", b"VAL1"),
            (basic, "VAL2", b"VAL2"),
            (basic, "a", ValueError),
            (basic, "val1", ValueError),
            (basic, None, ValueError),
            (basic_generator, "VAL1", b"VAL1"),
            (basic_generator, "VAL2", b"VAL2"),
            (basic_generator, "a", ValueError),
            (basic_generator, "val1", ValueError),
            (basic_generator, None, ValueError),
            (default, None, b"VAL1"),
            (default_optional, None, b"VAL1"),
            (optional, None, ValueError),
            (case_insensitive, "VAL1", b"VAL1"),
            (case_insensitive, "vAl2", b"vAl2"),
            (case_insensitive, "a", ValueError),
        ]

        self._unpack = [
            (basic, b"VAL1", "VAL1"),
            (basic, b"VAL2", "VAL2"),
            (basic, b"a", ValueError),
            (basic, None, ValueError),
            (default, None, "VAL1"),
            (default_optional, None, "VAL1"),
            (optional, None, None),
            (case_insensitive, b"val1", "val1"),
            (case_insensitive, b"vAl2", "vAl2"),
            (case_insensitive, b"a", ValueError),
        ]

    def test_discrete_values(self):
        values = ('VAL{}'.format(i + 1) for i in range(2))
        basic = Discrete(values)
        self.assertEqual(sorted(basic._values), sorted(basic._valid_values))


class TestLru(TestType):

    def setUp(self):
        basic = Lru()
        default = Lru(default=Lru.LRU_NOMINAL)
        optional = Lru(optional=True)
        default_optional = Lru(default=Lru.LRU_NOMINAL, optional=True)

        self._pack = [
            (basic, Lru.LRU_NOMINAL, b"nominal"),
            (basic, Lru.LRU_ERROR, b"error"),
            (basic, "a", ValueError),
            (basic, None, ValueError),
            (default, None, b"nominal"),
            (default_optional, None, b"nominal"),
            (optional, None, ValueError),
        ]

        self._unpack = [
            (basic, b"nominal", Lru.LRU_NOMINAL),
            (basic, b"error", Lru.LRU_ERROR),
            (basic, b"a", ValueError),
            (basic, None, ValueError),
            (default, None, Lru.LRU_NOMINAL),
            (default_optional, None, Lru.LRU_NOMINAL),
            (optional, None, None),
        ]


class TestAddress(TestType):

    def setUp(self):
        basic = Address()
        default = Address(default=("127.0.0.1", None))
        optional = Address(optional=True)
        default_optional = Address(default=("127.0.0.1", None), optional=True)

        self._pack = [
            (basic, ("127.0.0.1", None), b"127.0.0.1"),
            (basic, ("127.0.0.1", 80), b"127.0.0.1:80"),
            (basic, ("0:0:0:0:0:0:0:1", None), b"[0:0:0:0:0:0:0:1]"),
            (basic, ("::1", None), b"[::1]"),
            (basic, ("::FFFF:204.152.189.116", None),
             b"[::FFFF:204.152.189.116]"),
            (basic, ("::1", 80), b"[::1]:80"),
            (basic, "127.0.0.1", ValueError),  # value not a tuple
            (default, None, b"127.0.0.1"),
            (default_optional, None, b"127.0.0.1"),
            (optional, None, ValueError),
        ]

        self._unpack = [
            (basic, b"127.0.0.1", ("127.0.0.1", None)),
            (basic, b"127.0.0.1:80", ("127.0.0.1", 80)),
            (basic, b"[0:0:0:0:0:0:0:1]", ("0:0:0:0:0:0:0:1", None)),
            (basic, b"[::1]", ("::1", None)),
            (basic, b"[::FFFF:204.152.189.116]", ("::FFFF:204.152.189.116",
                                                 None)),
            (basic, b"[::1]:80", ("::1", 80)),
            (basic, b"127.0.0.1:foo", ValueError),
            (basic, None, ValueError),
            (default, None, ("127.0.0.1", None)),
            (default_optional, None, ("127.0.0.1", None)),
            (optional, None, None),
        ]


class TestTimestamp(TestType):

    def setUp(self):
        basic = Timestamp()
        default = Timestamp(default=1235475793.0324881)
        optional = Timestamp(optional=True)
        default_optional = Timestamp(default=1235475793.0324881, optional=True)

        self._pack = [
            (basic, 1235475381.6966901, b"1235475381.696690"),
            (basic, "a", ValueError),
            (basic, None, ValueError),
            (default, None, b"1235475793.032488"),
            (default_optional, None, b"1235475793.032488"),
            (optional, None, ValueError),
        ]

        self._unpack = [
            (basic, b"1235475381.696", 1235475381.6960001),
            (basic, b"a", ValueError),
            (basic, None, ValueError),
            (default, None, 1235475793.0324881),
            (default_optional, None, 1235475793.0324881),
            (optional, None, None),
        ]


class TestStrictTimestamp(TestType):

    def setUp(self):
        basic = StrictTimestamp()
        default = StrictTimestamp(default=1235475793.03249)
        optional = StrictTimestamp(optional=True)
        default_optional = StrictTimestamp(default=1235475793.03249,
                                           optional=True)

        self._pack = [
            (basic, 1235475381.69669, b"1235475381.69669"),
            (basic, -0.1, ValueError),
            (basic, -1.0, ValueError),
            (basic, "a", ValueError),
            (basic, None, ValueError),
            (default, None, b"1235475793.03249"),
            (default_optional, None, b"1235475793.03249"),
            (optional, None, ValueError),
        ]

        self._unpack = [
            (basic, b"1235475381.696", 1235475381.6960001),
            (basic, b"-0.1", ValueError),
            (basic, b"-1.0", ValueError),
            (basic, b"a", ValueError),
            (basic, None, ValueError),
            (default, None, 1235475793.03249),
            (default_optional, None, 1235475793.03249),
            (optional, None, None),
        ]


class TestTimestampOrNow(TestType):

    def setUp(self):
        basic = TimestampOrNow()
        default = TimestampOrNow(default=1235475793.0324881)
        optional = TimestampOrNow(optional=True)
        default_optional = TimestampOrNow(default=1235475793.0324881,
                                          optional=True)
        default_now = TimestampOrNow(default=TimestampOrNow.NOW)

        self._pack = [
            (basic, 1235475381.6966901, b"1235475381.696690"),
            (basic, "a", ValueError),
            (basic, TimestampOrNow.NOW, b"now"),
            (basic, None, ValueError),
            (default, None, b"1235475793.032488"),
            (default, TimestampOrNow.NOW, b"now"),
            (default_optional, None, b"1235475793.032488"),
            (optional, None, ValueError),
            (default_now, None, b"now"),
        ]

        self._unpack = [
            (basic, b"1235475381.696", 1235475381.6960001),
            (basic, b"a", ValueError),
            (basic, b"now", TimestampOrNow.NOW),
            (basic, None, ValueError),
            (default, None, 1235475793.0324881),
            (default, b"now", TimestampOrNow.NOW),
            (default_optional, None, 1235475793.0324881),
            (optional, None, None),
            (default_now, None, TimestampOrNow.NOW),
        ]


class TestStr(TestType):

    def setUp(self):
        basic = Str()
        default = Str(default="something")
        optional = Str(optional=True)
        default_optional = Str(default="something", optional=True)

        # For packing, the input can be any type except NoneType.
        self._pack = [
            (basic, "adsasdasd", b"adsasdasd"),
            (basic, b"adsasdasd", b"adsasdasd"),
            (basic, u"adsasdasd", b"adsasdasd"),
            (basic, u"skräm", b"skr\xc3\xa4m"),
            (basic, [1, 2.0, 'three', False], b"[1, 2.0, 'three', False]"),
            (basic, None, ValueError),
            (default, None, b"something"),
            (default_optional, None, b"something"),
            (optional, None, ValueError),
        ]

        # For unpacking, the input is assumed to be a byte string, and
        # the output should be a native string (bytes on PY2, unicode on PY3).
        # On PY2, as input is assumed to be bytes already, nothing is done,
        # nor is the type checked.
        if future.utils.PY2:
            bytes_unpacked = "adsasdasd"
            unicode_unpacked = u"adsasdasd"
            utf8_unpacked = "skr\xc3\xa4m"
            list_str_unpacked = "[1, 2.0, 'three', False]"
        else:
            bytes_unpacked = "adsasdasd"
            unicode_unpacked = AttributeError
            utf8_unpacked = "skräm"
            list_str_unpacked = "[1, 2.0, 'three', False]"

        self._unpack = [
            (basic, b"adsasdasd", bytes_unpacked),
            (basic, u"adsasdasd", unicode_unpacked),
            (basic, b"skr\xc3\xa4m", utf8_unpacked),
            (basic, b"[1, 2.0, 'three', False]", list_str_unpacked),
            (basic, None, ValueError),
            (default, None, "something"),
            (default_optional, None, "something"),
            (optional, None, None),
        ]


class TestStruct(TestType):

    def setUp(self):
        basic = Struct(">isf")
        default = Struct(">isf", default=(1, b"f", 3.4))
        optional = Struct(">isf", optional=True)
        default_optional = Struct(">isf", default=(1, b"f", 3.4), optional=True)

        self._pack = [
            (basic, (5, b"s", 2.5), b"\x00\x00\x00\x05s@ \x00\x00"),
            (basic, (b"s", 5, 2.5), ValueError),
            (basic, (5, b"s"), ValueError),
            (basic, None, ValueError),
            (default, None, b"\x00\x00\x00\x01f@Y\x99\x9a"),
            (default_optional, None, b"\x00\x00\x00\x01f@Y\x99\x9a"),
            (optional, None, ValueError),
        ]

        self._unpack = [
            (basic, b"\x00\x00\x00\x05s@ \x00\x00", (5, b"s", 2.5)),
            (basic, b"asdfgasdfas", ValueError),
            (basic, None, ValueError),
            (default, None, (1, b"f", 3.4)),
            (default_optional, None, (1, b"f", 3.4)),
            (optional, None, None),
        ]


class TestRegex(TestType):

    def setUp(self):
        basic = Regex("\d\d:\d\d:\d\d")
        default = Regex("\d\d:\d\d:\d\d", default="00:00:00")
        optional = Regex("\d\d:\d\d:\d\d", optional=True)
        default_optional = Regex("\d\d:\d\d:\d\d", default="00:00:00",
                                 optional=True)

        self._pack = [
            (basic, "12:34:56", b"12:34:56"),
            (basic, "sdfasdfsadf", ValueError),
            (basic, None, ValueError),
            (default, None, b"00:00:00"),
            (default_optional, None, b"00:00:00"),
            (optional, None, ValueError),
        ]

        self._unpack = [
            (basic, b"12:34:56", "12:34:56"),
            (basic, b"sdfasdfsadf", ValueError),
            (basic, None, ValueError),
            (default, None, "00:00:00"),
            (default_optional, None, "00:00:00"),
            (optional, None, None),
        ]


class TestDiscreteMulti(TestType):

    def setUp(self):
        basic = DiscreteMulti(("VAL1", "VAL2"))
        default = DiscreteMulti(("VAL1", "VAL2"), default=["VAL1"])
        optional = DiscreteMulti(("VAL1", "VAL2"), optional=True)
        default_optional = DiscreteMulti(("VAL1", "VAL2"), default=["VAL1"],
                                         optional=True)
        case_insensitive = DiscreteMulti(("val1", "VAL2"),
                                         case_insensitive=True)

        self._pack = [
            (basic, ["VAL1"], b"VAL1"),
            (basic, ["VAL2"], b"VAL2"),
            (basic, ["VAL1", "VAL2"], b"VAL1,VAL2"),
            (basic, "a", ValueError),
            (basic, "VAL1", ValueError),
            (basic, ["aaa"], ValueError),
            (basic, ["val1"], ValueError),
            (basic, ["VAL1", "val2"], ValueError),
            (basic, ["VAL1", "aaa"], ValueError),
            (basic, None, ValueError),
            (default, None, b"VAL1"),
            (default_optional, None, b"VAL1"),
            (optional, None, ValueError),
            (case_insensitive, ["VAL1"], b"VAL1"),
            (case_insensitive, ["vAl2"], b"vAl2"),
            (case_insensitive, ["VAL1", "val2"], b"VAL1,val2"),
            (case_insensitive, ["aaa"], ValueError),
        ]

        self._unpack = [
            (basic, b"VAL1", ["VAL1"]),
            (basic, b"VAL2", ["VAL2"]),
            (basic, b"VAL1,VAL2", ["VAL1", "VAL2"]),
            (basic, b"all", ["VAL1", "VAL2"]),
            (basic, b"VAL1,aaa", ValueError),
            (basic, b"VAL1,val2", ValueError),
            (basic, b"a", ValueError),
            (basic, None, ValueError),
            (default, None, ["VAL1"]),
            (default_optional, None, ["VAL1"]),
            (optional, None, None),
            (case_insensitive, b"val1", ["val1"]),
            (case_insensitive, b"vAl2", ["vAl2"]),
            (case_insensitive, b"VAL1,val2", ["VAL1", "val2"]),
            (case_insensitive, b"VAL1,aaa", ValueError),
            (case_insensitive, b"a", ValueError),
        ]


class TestDevice(object):
    def __init__(self):
        self.sent_messages = []

    @request(Int(min=1, max=10), Discrete(("on", "off")), Bool())
    @return_reply(Int(min=1, max=10), Discrete(("on", "off")), Bool())
    def request_one(self, req, i, d, b):
        if i == 3:
            return ("fail", "I failed!")
        if i == 5:
            return ("bananas", "This should never be sent")
        if i == 6:
            return ("ok", i, d, b, "extra parameter")
        if i == 9:
            self.finish_request_one(req, i, d, b)
            raise AsyncReply()
        return ("ok", i, d, b)

    @send_reply(Int(min=1, max=10), Discrete(("on", "off")), Bool())
    def finish_request_one(self, req, i, d, b):
        return (req, "ok", i, d, b)

    def reply(self, req, msg, orig_msg):
        self.sent_messages.append([req, msg])

    @request(Int(min=1, max=3, default=2),
             Discrete(("on", "off"), default="off"), Bool(default=True))
    @return_reply(Int(min=1, max=3), Discrete(("on", "off")), Bool())
    def request_two(self, req, i, d, b):
        return ("ok", i, d, b)

    @return_reply(Int(min=1, max=3), Discrete(("on", "off")), Bool())
    @request(Int(min=1, max=3), Discrete(("on", "off")), Bool())
    def request_three(self, req, i, d, b):
        return ("ok", i, d, b)

    @return_reply()
    @request()
    def request_four(self, req):
        return ["ok"]

    @inform(Int(min=1, max=3), Discrete(("on", "off")), Bool())
    def inform_one(self, i, d, b):
        pass

    @request(Timestamp(), Timestamp(optional=True), major=4)
    @return_reply(Timestamp(), Timestamp(default=321), major=4)
    def request_katcpv4_time(self, req, timestamp1, timestamp2):
        self.katcpv4_time1 = timestamp1
        self.katcpv4_time2 = timestamp2
        if timestamp2:
            return ('ok', timestamp1, timestamp2)
        else:
            return ('ok', timestamp1)

    @request(Timestamp(multiple=True), major=4)
    @return_reply(Timestamp(multiple=True), major=4)
    def request_katcpv4_time_multi(self, req, *timestamps):
        self.katcpv4_time_multi = timestamps
        return ('ok',) + timestamps

    @return_reply(Int(), Str())
    @request(Int(), include_msg=True)
    def request_eight(self, req, msg, i):
        return ("ok", i, msg.name)

    @request(Int(), Float(multiple=True))
    @return_reply(Int(), Float(multiple=True))
    def request_int_multifloat(self, req, i, *floats):
        return ('ok', i) + floats

class TestDecorator(unittest.TestCase):
    def setUp(self):
        self.device = TestDevice()

    def test_request_multi(self):
        with self.assertRaises(TypeError) as ex:
            request(Bool(multiple=True), Int())
        self.assertEqual(
            str(ex.exception),
            'Only the last parameter type can accept multiple arguments.')

    def test_return_reply_multi(self):
        with self.assertRaises(TypeError) as ex:
            return_reply(Bool(multiple=True), Int())
        self.assertEqual(
            str(ex.exception),
            'Only the last parameter type can accept multiple arguments.')

    def test_katcpv4(self):
        ts = 12345678                     # In milliseconds
        req = mock.Mock()
        ret_msg = self.device.request_katcpv4_time(req, Message.request(
            'katcpv4-time', str(ts)))
        self.assertTrue(ret_msg.reply_ok())
        self.assertAlmostEqual(float(ret_msg.arguments[1]), ts)
        # Test decorator default value
        self.assertAlmostEqual(float(ret_msg.arguments[2]), 321*SEC_TO_MS_FAC)
        self.assertAlmostEqual(self.device.katcpv4_time1*SEC_TO_MS_FAC, ts)
        self.assertEqual(self.device.katcpv4_time2, None)
        ts1 = 1234
        ts2 = 2345
        ret_msg = self.device.request_katcpv4_time(req, Message.request(
            'katcpv4-time', str(ts1), str(ts2)))
        self.assertTrue(ret_msg.reply_ok())
        self.assertAlmostEqual(float(ret_msg.arguments[1]), ts1)
        self.assertAlmostEqual(float(ret_msg.arguments[2]), ts2)
        self.assertAlmostEqual(self.device.katcpv4_time1*SEC_TO_MS_FAC, ts1)
        self.assertAlmostEqual(self.device.katcpv4_time2*SEC_TO_MS_FAC, ts2)

    def test_katcpv4_multi(self):
        tss = (1234, 5678, 9012)                     # In milliseconds
        req = mock.Mock()
        ret_msg = self.device.request_katcpv4_time_multi(req, Message.request(
            'katcpv4-time-multi', *(str(ts) for ts in tss) ))
        for i, ts in enumerate(tss):
            self.assertAlmostEqual(float(ret_msg.arguments[i+1]), ts)
            self.assertAlmostEqual(self.device.katcpv4_time_multi[i],
                                   ts*MS_TO_SEC_FAC)

    def test_request_one(self):
        """Test request with no defaults."""
        req = mock.Mock()
        req.msg.name = 'one'
        self.assertEqual(str(self.device.request_one(req, Message.request(
                        "one", "2", "on", "0"))), "!one ok 2 on 0")
        self.assertRaises(FailReply, self.device.request_one, req,
                          Message.request("one", "14", "on", "0"))
        self.assertRaises(FailReply, self.device.request_one, req,
                          Message.request("one", "2", "dsfg", "0"))
        self.assertRaises(FailReply, self.device.request_one, req,
                          Message.request("one", "2", "on", "3"))
        self.assertRaises(FailReply, self.device.request_one, req,
                          Message.request("one", "2", "on", "0", "3"))

        self.assertRaises(FailReply, self.device.request_one, req,
                          Message.request("one", "2", "on"))

        self.assertEqual(str(self.device.request_one(req, Message.request(
                        "one", "3", "on", "0"))), "!one fail I\\_failed!")
        self.assertRaises(ValueError, self.device.request_one, req,
                          Message.request("one", "5", "on", "0"))
        self.assertRaises(ValueError, self.device.request_one, req,
                          Message.request("one", "6", "on", "0"))

        req.reset_mock()
        self.assertRaises(AsyncReply, self.device.request_one, req,
                          Message.request("one", "9", "on", "0"))
        self.assertEqual(req.reply_with_message.call_count, 1)
        req.reply_with_message.assert_called_once_with(Message.reply(
            'one', 'ok', '9', 'on', '0'))

    def test_request_two(self):
        """Test request with defaults."""
        req = ""
        self.assertEqual(str(self.device.request_two(req, Message.request(
                        "two", "2", "on", "0"))), "!two ok 2 on 0")
        self.assertRaises(FailReply, self.device.request_two, req,
                          Message.request("two", "4", "on", "0"))
        self.assertRaises(FailReply, self.device.request_two, req,
                          Message.request("two", "2", "dsfg", "0"))
        self.assertRaises(FailReply, self.device.request_two, req,
                          Message.request("two", "2", "on", "3"))

        self.assertEqual(str(self.device.request_two(req, Message.request(
                        "two", "2", "on"))), "!two ok 2 on 1")
        self.assertEqual(str(self.device.request_two(req, Message.request(
                        "two", "2"))), "!two ok 2 off 1")
        self.assertEqual(str(self.device.request_two(req, Message.request(
                        "two"))), "!two ok 2 off 1")

    def test_request_three(self):
        """Test request with no defaults and decorators in reverse order."""
        req = ""
        self.assertEqual(str(self.device.request_three(req, Message.request(
                        "three", "2", "on", "0"))), "!three ok 2 on 0")
        self.assertRaises(FailReply, self.device.request_three, req,
                          Message.request("three", "4", "on", "0"))
        self.assertRaises(FailReply, self.device.request_three, req,
                          Message.request("three", "2", "dsfg", "0"))
        self.assertRaises(FailReply, self.device.request_three, req,
                          Message.request("three", "2", "on", "3"))

        self.assertRaises(FailReply, self.device.request_three, req,
                          Message.request("three", "2", "on"))

    def test_request_four(self):
        """Test request with no defaults and no parameters or return values"""
        req = ""
        self.assertEqual(str(self.device.request_four(req, Message.request(
                        "four"))), "!four ok")

    def test_inform_one(self):
        """Test inform with no defaults."""
        req = ""
        self.assertEqual(self.device.inform_one(Message.inform(
                         "one", "2", "on", "0")), None)

    def test_request_eight(self):
        """Test server request with a message argument."""
        req = ""
        self.assertEqual(str(self.device.request_eight(req, Message.request(
                         "eight", "8"))), "!eight ok 8 eight")

    def test_request_int_multifloat(self):
        req = self.device.request_int_multifloat
        desired_i, desired_floats = (7, (1.2, 999, 71.43))
        self.assertEqual(str(req('req', Message.request(
            'int-multifloat', desired_i, *desired_floats))),
                         '!int-multifloat ok 7 1.2 999.0 71.43')
        with self.assertRaises(FailReply) as ex:
            req('req', Message.request('int-multifloat', desired_i, 1.2, 'abc'))
        # storing Message.arguments as byte string results in slightly different
        # reprs for PY2 compared to PY3.
        if future.utils.PY2:
            expected = "Error in parameter 3 (): Could not parse value 'abc' as float."
        else:
            expected = "Error in parameter 3 (): Could not parse value 'b'abc'' as float."
        self.assertEqual(str(ex.exception), expected)


class TestPackTypes(unittest.TestCase):

    def check_packing(self, types, args, expected):
        list_packed = pack_types(list(types), list(args), DEFAULT_KATCP_MAJOR)
        tuple_packed = pack_types(tuple(types), tuple(args), DEFAULT_KATCP_MAJOR)
        self.assertEqual(list_packed, expected)
        self.assertEqual(tuple_packed, expected)

    def test_pack_types_empty(self):
        expected = []
        self.check_packing([], [], expected)

    def test_pack_types_single(self):
        expected = [b'one']
        self.check_packing([Str()], ['one'], expected)

    def test_pack_types_many_without_multiple(self):
        expected = [b'one', b'2']
        self.check_packing([Str(), Int()], ['one', 2], expected)

    def test_pack_types_many_with_multiple(self):
        expected = [b'one', b'2', b'3']
        self.check_packing([Str(), Int(multiple=True)], ['one', 2, 3], expected)

    def test_pack_types_more_types_than_args(self):
        expected = [b'one', b'2', b'1', b'four']
        self.check_packing(
            [Str(), Int(), Bool(default=True), Str(default='four')],
            ['one', 2],
            expected)

    def test_pack_types_more_args_than_types_fails(self):
        with self.assertRaises(ValueError):
            self.check_packing([], ['one'], [])


class TestUnpackTypes(unittest.TestCase):

    def check_unpacking(self, types, args, expected):
        arg_names = ['arg'] * len(args)  # names don't matter
        list_unpacked = unpack_types(
            list(types), list(args), arg_names, DEFAULT_KATCP_MAJOR)
        tuple_unpacked = unpack_types(
            tuple(types), tuple(args), tuple(arg_names), DEFAULT_KATCP_MAJOR)
        self.assertEqual(list_unpacked, expected)
        self.assertEqual(tuple_unpacked, expected)

    def test_unpack_types_empty(self):
        expected = []
        self.check_unpacking([], [], expected)

    def test_unpack_types_single(self):
        expected = ['one']
        self.check_unpacking([Str()], [b'one'], expected)

    def test_unpack_types_many_without_multiple(self):
        expected = ['one', 2]
        self.check_unpacking([Str(), Int()], [b'one', b'2'], expected)

    def test_unpack_types_many_with_multiple(self):
        expected = ['one', 2, 3]
        self.check_unpacking([Str(), Int(multiple=True)], [b'one', b'2', b'3'], expected)

    def test_unpack_types_more_types_than_args(self):
        expected = ['one', 2, True, None]
        self.check_unpacking(
            [Str(), Int(), Bool(default=True), Str(optional=True)],
            [b'one', b'2'],
            expected)

    def test_unpack_types_more_args_than_types_fails(self):
        with self.assertRaises(FailReply):
            self.check_unpacking([], ['one'], [])
